# A Multi-user "Chat" system

# =============================================================
# Do some standard groupkit setups...

# Initialize conference
gk_initConf $argv

# Tell the window manager the title and the default window size
wm title . "Multi-User Chat"
wm minsize . 50 6

# Create the default groupkit pulldown menu bar 
gk_defaultMenu .menubar
pack .menubar -side top -fill x 

# =============================================================
# Handle arrival, departure, and updating of participants 

# A new person has arrived. Create a chat window for them
trace variable gk_newUser w makeChatWindow
proc makeChatWindow args { global gk_newUser
    createChatWindow \
	[keylget gk_newUser usernum] \
	[keylget gk_newUser username] \
	disabled
    return ""
}

# A person has left. Remove their chat window 
trace variable gk_deletedUser w removeChatWindow
proc removeChatWindow args { global gk_deletedUser 
    set usernum [keylget gk_deletedUser usernum]
    catch {destroy .$usernum}
    return ""
}

# A person has joined. It is our responsibility to update them
# Only one of the replicated processes will actually do this...
trace variable gk_entrantUserNum w {updateEntrant}
proc updateEntrant args { global gk_entrantUserNum _gk_localUser
    set frames [winfo children .]
    foreach i $frames {
	# Look for the other windows...
	if {$i != ".menubar" && [string compare $i .$gk_entrantUserNum] } {
	    gk_toUserNum $gk_entrantUserNum \
		update_text_window $i [$i.f.text get 0.0 end]
	}
    }
}

# Given a window and some text, insert the text in the window
proc update_text_window {frame text} {
    if {[winfo exists $frame.f.text]} {
	$frame.f.text configure -state normal
	$frame.f.text insert 0.0 $text
	$frame.f.text configure -state disabled
    } else {
	# We have to delay a bit, because the application is
	# still trying to make the window!
	puts "Nope, try again"
	after 500 "update_text_window $frame $text"
    }
}

# Actually create the chat window. The state says whether it
# it should be editable or not by the user.
proc createChatWindow {number name state} {
    set chatframe .$number 
    set label .$number.label
    set textframe .$number.f
    set text  .$number.f.text 
    set scroll  .$number.f.scroll

    frame $chatframe -bd 2
    pack $chatframe -side top -expand yes -fill both

    label $label \
	-text $name \
	-font -Adobe-times-bold-r-normal--*-120*
    pack  $label -side top -anchor w

    frame $textframe 

    text  $text \
	-state $state \
	-relief ridge \
	-bd 2 \
	-width 50 -height 6 \
	-wrap word \
	-setgrid true \
	-yscrollcommand "$scroll set"

    scrollbar $scroll \
	-orient vertical \
	-relief ridge \
	-command "$text yview"

    pack  $textframe -side top -expand yes -fill both
    pack  $text -side left -fill both -expand yes
    pack  $scroll -side right -fill y
    if {$state != "disabled" } {
	focus $text
	set return "\n"
	bind  $text  <Any-KeyPress> "insertCharacter $number %A"
    }
}

# ================================================================
# These handlers routines take care of inserting the characters
# the user had typed in the proper place and the proper text widget

proc insertCharacter {number char} {
    set position [.$number.f.text index insert]
    gk_toAll reallyInsertCharacter $number $char $position
}

# We over-ride all key bindings. Thus we handle all 
# character interpretation ourselves.
proc reallyInsertCharacter {number char position} {
    global _gk_localUser 
    set w .$number.f.text

    # If its disabled, make it normal so we can maniplate it
    if {$number != [keylget _gk_localUser usernum]} {
	$w configure -state normal
    }
    case $char in {
	{\012 \015} {	    
	    # Returns, Line feeds
	    $w insert $position "\n"
	    $w yview -pickplace insert
	} 
	{\010 \177} {
	    # Backspace
            $w mark set insert $position
	    tk_textBackspace $w; 
	    $w yview -pickplace insert
	}
	{\004} {
	    # Forward delete character
	    set old [$w index insert]
            $w mark set insert $position+1c
	    if {$old !=  [$w index insert]} {
		tk_textBackspace $w; 
		$w yview -pickplace insert
	    }
	}
	{\006} {
	    # Forward Character (Control-f)
	    $w mark set insert $position+1c
	    $w yview -pickplace insert
	}
	{\002} {       
	    # Backward Character (Control-b)
	    $w mark set insert $position-1c
	    $w yview -pickplace insert
	}
	{\001} {
	    # Beginning of line (Control-e)            
	    $w mark set insert "insert linestart"
	    $w yview -pickplace insert
	}
	{\005} {
	    # End of line (Control-e)            
	    $w mark set insert "insert lineend"
	    $w yview -pickplace insert
	}
	{\016} {            
	    # Previous line (Control-p)
	    $w mark set insert "insert +1l"
	    $w yview -pickplace insert
	}
	{\020} {            
	    # Next line (Control-n)
	    $w mark set insert "insert-1l"
	    $w yview -pickplace insert
	}
	default {
	    # Insert printing characters only
	    if {$char >= " " && $char <= "~"} {
		$w insert $position $char
		$w yview -pickplace $position
	    }
	}
    }
    # All windows but the user's should be disabled
    if {$number != [keylget _gk_localUser usernum]} {
	$w configure -state disabled
    }
}


# ================================================================
# And lets not forget to make my own chat window!
set me [keylget _gk_localUser username] 
createChatWindow [keylget _gk_localUser usernum] "I'm $me" normal

# ================================================================
# These variables/routines are used to describe this application
# to the user. An "About this application" item is added to the
# Help menu, which will pop up an information window

set application(title) \
    "A Multi-User Chat System"

set application(description) \
"You can have real-time text conversations with other\
conference participants.

Every person has their own window. You will see everything they type\
in their window, and vice-versa.

Emacs-like key bindings:
Control-f  Forward character
Control-b  Backward character
Control-a  Beginning of line
Control-e  End of line
Control-p  Previous line
Control-n  Next line
Control-d  Delete forward character
Control-h  Backspace
"

.menubar.help.menu add command \
    -label "About This Application" \
    -command {about_this_applicaton}


# Create a window that describes this application to the user
proc about_this_applicaton {} {global application
    if {[winfo exists .about_window] == 1} { return }

    toplevel .about_window 
    wm title .about_window "About this application"

    label .about_window.title \
	-text $application(title) \
	-font -Adobe-times-bold-r-normal--*-180*

    message .about_window.text  -width 240 \
	-text $application(description)

    button .about_window.ok -text "Okay" -command "destroy .about_window"

    pack .about_window.title .about_window.text .about_window.ok -side top
}



