#
# Open Registration Client
# ------------------------
#
# This module presents one possible Registrar Client (see rc.tcl for 
# information on Registrar Clients in general).  Within an open Registrar
# Client, any user can create a conference, and users can join any 
# conference.  Users can only delete themselves from the conference,
# and the last user leaving a conference closes down the entire conference
# (default behavior for Registrar Clients; see the gk_userLeft proc in rc.tcl).
#
# There are two parts to this module, the underlying protocol part, and a
# Tk interface to interact with the protocol.
#

_gk_initGenericRC

_gk_pollConferences

#
# This routine allows a user to join a conference and is called in 
#    response to double clicking an existing conference in the interface
#
# This routine tells the Registrar we want to join the conference, and then
# instructs our conference (which hasn't yet been created; the _gk_joinTo proc
# calls _gk_toConf which will queue the messages until it is created) to join to
# join to the existing users of the conference.  The global list "_gk_confsJoined"
# keeps track of conferences we've already joined, so as not to join them
# twice.
#
# Note: we no longer join existing conferences; this is moved to gk_foundNewUser
#

set _gk_confsJoined ""

proc _gk_doJoin whichconf {     global _gk_rcConfsList _gk_confsJoined conferences
    set conf [lindex $_gk_rcConfsList $whichconf]
    if {([info exists conferences] == 1) && ($conferences != "")} {
	foreach temp $conferences {
	    if {[keylget temp confnum] == [keylget conf confnum]} { return }
	}
    }
    if {[lsearch $_gk_confsJoined [keylget conf confnum]] == -1} {
	lappend _gk_confsJoined [keylget conf confnum]
	_gk_callJoinConference [keylget conf confnum]
	after 100 _gk_pollUsers [keylget conf confnum]
    }
}


#
# This is called by _gk_updateList when a new conference is found.  Add the
# conference to the global _gk_rcConfsList.  If we created the conference, as
# specified by the originator key, tell the Registrar we'd like to join
# to it; under this policy we automatically join to conferences we create
#

proc gk_foundNewConf conf {     global _gk_rcConfsList host myport _gk_rcUsersList
    _gk_pollUsers [keylget conf confnum]
    lappend _gk_rcConfsList $conf
    set _gk_rcUsersList([keylget conf confnum]) {}
    if {[keylget conf originator] == [concat $host$myport]} {
	_gk_callJoinConference [keylget conf confnum]
	_gk_pollUsers [keylget conf confnum]
    }
}


#
# This is called by _gk_updateList when a new user is found.  Add the user to 
# the appropriate user list.  If the new user is in fact ourself, create
# the conference application.
#
# Note: when we get our own entry, we also join to other users in the conference.
#   Because joins are essentially serialized through the registrar, after we
#   receive our own "newuser" announcement, there will not be any announcements
#   for "previously joined" users.
#

proc gk_foundNewUser user {    global host myport _gk_rcUsersList
    set confnum [keylget user confnum]
    lappend _gk_rcUsersList($confnum) $user
    if {([keylget user host] == $host) && ([keylget user port] == $myport)} {
	_gk_createConference [_gk_findConfInfo $confnum] [keylget user usernum]
	foreach i $_gk_rcUsersList($confnum) {
	    if {[keylget i usernum] < [keylget user usernum]} {
		_gk_joinTo $i $user
	    }
	}
    }
}


#
# This is called by _gk_updateList when a deleted user is found.  Delete the
# user from the list and tell the conference to remove the user which may
# even be ourself (because under the protocol we obey all delete requests).
# If the user is ourself, delete the conference from the _gk_confsJoined list
# so we can join it again if desired.
#

proc gk_foundDeletedUser {conf user} {global _gk_rcUsersList _gk_confsJoined \
				host myport 
#				host myport _gk_stillCheck


# kill the _gk_confsJoined entry if deleted user was us - before we do
#   ANYTHING else

    foreach i $_gk_rcUsersList($conf) {
	if {[keylget i usernum] == $user} {
	    if {([keylget i host] == $host) && ([keylget i port] == $myport)} {
# Don't check connection---user is leaving.
#		set _gk_stillCheck no
	        set idx [lsearch $_gk_confsJoined $conf]
	        if {$idx != -1} { 
		    set _gk_confsJoined [lreplace $_gk_confsJoined $idx $idx]
	        }
	    }
	}
    }	

    set idx 0; while {$idx < [llength $_gk_rcUsersList($conf)]} {
	set i [lindex $_gk_rcUsersList($conf) $idx]
	if {[keylget i usernum] == $user} {
	    catch {_gk_toConf $conf "_gk_removeUser [list [lindex $_gk_rcUsersList($conf) $idx]]"}
	    set _gk_rcUsersList($conf) [lreplace $_gk_rcUsersList($conf) $idx $idx]
	} else { incr idx }
    }

    _gk_updateUserList blah
}


#
# This is called by _gk_updateList when a deleted conference is found.  Delete
# it from the list and tell the conference to delete itself (if we are
# running the conference).
#

proc gk_foundDeletedConf conf { global _gk_rcConfsList _gk_rcUsersList
    set idx 0; foreach i $_gk_rcConfsList {
	if {[keylget i confnum] == $conf} {
	    set _gk_rcConfsList [lreplace $_gk_rcConfsList $idx $idx]
	    set _gk_rcUsersList($conf) ""
	    _gk_toConf $conf "_gk_deleteConf"
	} else { incr idx }
    }
}


#
# The remainder of this module specifies the Tk interface for the Open
# Registrar Client.  
#
# This first part specifies the main window.  It contains two scrollable
# lists, one containing a list of all the conferences (their names actually)
# and the second a list of users for a particular conference (selected in
# the first list).  Double clicking on a conference allows you to join
# the conference.  A 'new' button pops up a dialog box allowing you to
# create new conferences.
#

pack append [frame .buttons -borderwidth 10] \
    [button    .buttons.new -text New -command "_gk_newConfDlg"] {left expand}\

pack append [frame .conflist] \
    [label     .conflist.label -text Conferences] top \
    [scrollbar .conflist.scroll -relief sunken \
	       -command ".conflist.list yview"] {right filly} \
    [listbox   .conflist.list -yscroll ".conflist.scroll set" \
	       -relief sunken -exportselection false -setgrid 1] {left expand fill}

pack append [frame .userlist] \
    [label     .userlist.label -text Users] top \
    [scrollbar .userlist.scroll -relief sunken  \
               -command ".userlist.list yview"] {right filly} \
    [listbox   .userlist.list -yscroll ".userlist.scroll set" \
               -relief sunken -exportselection false -setgrid 1] {left expand fill}

gk_defaultRCMenu .menu

pack append . \
    .menu {top fillx} \
    .buttons {bottom expand fill} \
    .conflist left \
    .userlist right

bind .conflist.list <1> {%W select from [%W nearest %y]; _gk_updateUserList blat}
bind .conflist.list <Double-1> {_gk_doJoin [.conflist.list curselection]}
tk_listboxSingleSelect .conflist.list


#
# The interface monitors both the _gk_rcUsersList and the _gk_rcConfsList, and when they
# change calls routines to update the interface.
#

trace variable _gk_rcConfsList w _gk_updateConfList 
trace variable _gk_rcUsersList w _gk_updateUserList


#
# When the conference list changes, rebuild the entire conferences listbox.
# Not overly efficient, but hey, it works.
#

proc _gk_updateConfList args {
    global _gk_rcConfsList .conflist.list
    .conflist.list delete 0 end
    foreach i $_gk_rcConfsList {
	set blah $i
	.conflist.list insert end [keylget blah confname]
    }
    return
}


#
# When the user list changes, update the user listbox.  This listbox displays
# the users of the conference selected in the conferences listbox.
#

proc _gk_updateUserList args {     global _gk_rcConfsList _gk_rcUsersList
    set sel [.conflist.list curselection]
    .userlist.list delete 0 end
    if {$sel != ""} {
	set conf [lindex $_gk_rcConfsList $sel]
	set confnum [keylget conf confnum]
	foreach i $_gk_rcUsersList($confnum) {
	    set blah $i
	    .userlist.list insert end [keylget blah userid]
	}
    }
    return
}


#
# Pop up a dialog box allowing the user to create a new conference.  The
# box prompts for the name of the dialog box as well as the type of conference
# to be created.  The list of conference types is obtained from the .Xdefaults
# file (see the documentation for proc _gk_createConference in rc.tcl for the
# format of the particular resources used).
#
# Changed to support the gk_program tcl variable rather than the
# Xdefaults form.
#

proc _gk_newConfDlg {} {    global conftype confName gk_program
    if {[info commands .dlg] == ".dlg"} return

    toplevel .dlg
    wm title .dlg "New Conference"
    set confName "Conference name"

    pack append [frame .dlg.buttons] \
	[button .dlg.buttons.cancel -text Cancel \
	        -command "destroy .dlg"] right \
	[button .dlg.buttons.ok -text Ok -command _gk_doNewConf] right

    frame .dlg.types -borderwidth 20
    
    set j 1; foreach i [array names gk_program] {
	if {$j==1} { set conftype [set i] }
	pack append .dlg.types \
	    [radiobutton .dlg.types.r$j -text [set i] -variable conftype -value [set i]] \
	    {top expand pady 5 fillx frame w}
	incr j
    }

    pack append .dlg \
	[label .dlg.lbl -text "Conference name"] top \
	[entry .dlg.name -relief sunken -textvariable confName] \
	           {top pady 10 fillx} \
	.dlg.buttons {bottom fillx pady 20} \
	.dlg.types top
}


#
# This command is called in response to pressing the "Ok" button in the
# new conferences dialog box.  Create the desired conference.
#

proc _gk_doNewConf {} {	global confName conftype host myport
    keylset conf confname $confName conftype $conftype originator $host$myport

    if {[string compare $confName "Conference name"] == 0} {
	keylset conf confname [keylget conf conftype]}
    _gk_callNewConference $conf
    _gk_pollConferences
    destroy .dlg
}

